/*
   strstr.S
   Search string contained in a string

   Contributed by:
     Philip Soeberg <philip@soeberg.net>
*/

#include "macros.inc"

#define s1_hi r25
#define s1_lo r24
#define s2_hi r23
#define s2_lo r22

#define chr1_lo r18     //first char of str1 (updated in loop)
#define chr1_hi r19     //second char of str1 (updated in loop)
#define chr2_lo r20     //first char of str2 (updated in loop)
#define chr2_hi r21     //second char of str2 (updated in loop)

#define ret_hi r25
#define ret_lo r24

/** \ingroup avr_string
    \fn char *strstr(const char *s1, const char *s2)
    \brief Locate a substring.

    The strstr() function finds the first occurrence of the substring \c s2
    in the string <tt>s1</tt>.  The terminating `\0' characters are not
    compared.

    \returns The strstr() function returns a pointer to the beginning of the
    substring, or NULL if the substring is not found. if s2 points to a string
    of zero length, the function returns s1. */

.text
.global _U(strstr)
.type _U(strstr), @function
_U(strstr):

        push r18        //Store temp registers
        push r19
        push r20
        push r21

        LOAD_X(s1_lo, s1_hi)
        LOAD_Z(s2_lo, s2_hi)
        ld chr2_lo, Z+
        tst chr2_lo             //is str2 empty?
        breq .strstr_empty_str2
        ld chr2_hi, Z+          //second char from str2
        tst chr2_hi             //is str2 a one-char string?
        breq .strstr_strchrcall //if so, use strchr code

        clr ret_lo
        clr ret_hi

.strstr_findnext:
        ld chr1_lo, X+
        cp  chr2_lo, chr1_lo
        breq .strstr_found_char

        tst chr2_lo
        breq .strstr_not_found

.strstr_start_loop:
        ld chr1_lo, X+              //put next char from str1 into chr1_lo

.strstr_loop:
        cp chr1_lo, chr2_lo
        breq .strstr_found_char

        tst chr1_lo                 //end of str1?
        brne .strstr_start_loop     //no, go get another char from str1

.strstr_not_found:
        clr ret_lo
        clr ret_hi
        rjmp .strstr_ret

.strstr_found_char:             //recall that chr2_hi contains second char from str2

        mov ret_lo, XL          //Store pointer to where match found @ str1
        mov ret_hi, XH

        ld chr1_lo, X+          //put next char from str1 into chr1_lo

        cp chr2_hi, chr1_lo     //compare second chars
        brne .strstr_loop       //no match, continue search

.strstr_compare_loop:
        ld chr2_hi, Z+          //put next char from str2 into chr2_hi
        tst chr2_hi             //end of str2?
        breq .strstr_match      //if so, then a match has been found

        ld chr1_lo, X+          //get next char from str1

        cp chr1_lo, chr2_hi     //are chars from str1 and str2 equal?
        brne .strstr_findnext   //no

        ld chr2_lo, Z+          //put the next char of str2 into chr2_lo
        tst chr2_lo             //end of str2?
        breq .strstr_match      //if so, then a match has been found

        ld chr1_hi, X+          //get next char from str1
        cp chr2_lo, chr1_hi     //are chars from str1 and str2 equal?
        breq .strstr_compare_loop

//no match. test some more chars (to improve execution time for bad strings)
        rjmp .strstr_findnext


.strstr_strchrcall:
        mov s2_lo, chr2_lo  //strchr expects a char, not a *char.
        call strchr
        rjmp .strstr_ret

.strstr_match:
        sbiw ret_lo, 1      //subtract 1 from str1 pointer
        rjmp .strstr_ret

.strstr_empty_str2:
        mov ret_lo, s1_lo   //Empty str2 string, return str1 (ANSI mandated!)
        mov ret_hi, s1_hi

.strstr_ret:
        pop r21             //Restore temp register
        pop r20
        pop r19
        pop r18
        ret

.strstr_end:
.size _U(strstr), .strstr_end - _U(strstr)
